/****************************************************************************
*   Squeak port for OS/2
*
*   - Named External Primitives support
*   - Both liked in Squeak VM and in DLL's
*
* Feel free to experiment with this files, and to modify them anyway you like.
* If you fix any bugs, or enhance something, please let me know. I will
* include any enhancement on future releases. Bug reports are also welcome.
* I compile Squeak with IBM's VisualAge C++ for OS/2, but I think you can
* use other compilers as well if they support multimedia APIs (MMPM2.LIB).
* Enyoy!
*
*   Juan Manuel Vuletich
*    jvuletic@dc.uba.ar
*
*****************************************************************************/
#define INCL_DOSMODULEMGR
#define INCL_DOSERRORS
#include <os2.h>
#include <stdio.h>
#include "sq.h"

// This is to include warnings to the user in case of API failures (in addition to primitive fail)
#define WARNINGS 1

/***************************************************************************/
/* Function Prototypes                                                     */
/*          Support Function from sqOS2Window.c                            */
/***************************************************************************/
char ErrorMessage[200];
void ShowErrorMessage(char* ErrorMessage);
void ShowWarningMessage(char* ErrorMessage);

/* Declaration for EXPORTed functions that are included into Squeak EXE
   You must also include them in getExternalFunctionInSqueakModule () */
EXPORT(int) gePrimitiveAbortProcessing(void);
EXPORT(int) gePrimitiveAddActiveEdgeEntry(void);
EXPORT(int) gePrimitiveAddBezier(void);
EXPORT(int) gePrimitiveAddBezierShape(void);
EXPORT(int) gePrimitiveAddBitmapFill(void);
EXPORT(int) gePrimitiveAddCompressedShape(void);
EXPORT(int) gePrimitiveAddGradientFill(void);
EXPORT(int) gePrimitiveAddLine(void);
EXPORT(int) gePrimitiveAddOval(void);
EXPORT(int) gePrimitiveAddPolygon(void);
EXPORT(int) gePrimitiveAddRect(void);
EXPORT(int) gePrimitiveChangedActiveEdgeEntry(void);
EXPORT(int) gePrimitiveCopyBuffer(void);
EXPORT(int) gePrimitiveDisplaySpanBuffer(void);
EXPORT(int) gePrimitiveDoProfileStats(void);
EXPORT(int) gePrimitiveFinishedProcessing(void);
EXPORT(int) gePrimitiveGetAALevel(void);
EXPORT(int) gePrimitiveGetBezierStats(void);
EXPORT(int) gePrimitiveGetClipRect(void);
EXPORT(int) gePrimitiveGetCounts(void);
EXPORT(int) gePrimitiveGetDepth(void);
EXPORT(int) gePrimitiveGetFailureReason(void);
EXPORT(int) gePrimitiveGetOffset(void);
EXPORT(int) gePrimitiveGetTimes(void);
EXPORT(int) gePrimitiveInitializeBuffer(void);
EXPORT(int) gePrimitiveInitializeProcessing(void);
EXPORT(int) gePrimitiveMergeFillFrom(void);
EXPORT(int) gePrimitiveNeedsFlush(void);
EXPORT(int) gePrimitiveNeedsFlushPut(void);
EXPORT(int) gePrimitiveNextActiveEdgeEntry(void);
EXPORT(int) gePrimitiveNextFillEntry(void);
EXPORT(int) gePrimitiveNextGlobalEdgeEntry(void);
EXPORT(int) gePrimitiveRegisterExternalEdge(void);
EXPORT(int) gePrimitiveRegisterExternalFill(void);
EXPORT(int) gePrimitiveRenderImage(void);
EXPORT(int) gePrimitiveRenderScanline(void);
EXPORT(int) gePrimitiveSetAALevel(void);
EXPORT(int) gePrimitiveSetClipRect(void);
EXPORT(int) gePrimitiveSetColorTransform(void);
EXPORT(int) gePrimitiveSetDepth(void);
EXPORT(int) gePrimitiveSetEdgeTransform(void);
EXPORT(int) gePrimitiveSetOffset(void);
EXPORT(int) m23PrimitiveComposeMatrix(void);
EXPORT(int) m23PrimitiveInvertPoint(void);
EXPORT(int) m23PrimitiveInvertRectInto(void);
EXPORT(int) m23PrimitiveIsIdentity(void);
EXPORT(int) m23PrimitiveIsPureTranslation(void);
EXPORT(int) m23PrimitiveTransformPoint(void);
EXPORT(int) m23PrimitiveTransformRectInto(void);
EXPORT(int) primitiveFFTPermuteData(void);
EXPORT(int) primitiveFFTScaleData(void);
EXPORT(int) primitiveFFTTransformData(void);
EXPORT(int) primitiveFloatArrayAddFloatArray(void);
EXPORT(int) primitiveFloatArrayAddScalar(void);
EXPORT(int) primitiveFloatArrayAt(void);
EXPORT(int) primitiveFloatArrayAtPut(void);
EXPORT(int) primitiveFloatArrayDivFloatArray(void);
EXPORT(int) primitiveFloatArrayDivScalar(void);
EXPORT(int) primitiveFloatArrayEqual(void);
EXPORT(int) primitiveFloatArrayHash(void);
EXPORT(int) primitiveFloatArrayMulFloatArray(void);
EXPORT(int) primitiveFloatArrayMulScalar(void);
EXPORT(int) primitiveFloatArraySubFloatArray(void);
EXPORT(int) primitiveFloatArraySubScalar(void);
EXPORT(int) setInterpreter(int anInterpreter);

typedef struct ModuleEntry {
   struct ModuleEntry *next;
   char name[255];
   HMODULE handle;
} ModuleEntry;

static ModuleEntry *firstEntry = NULL;

/* In this module */
void* getExternalFunctionInSqueakModule(char* functionName);

/* Find the handle for the given module */
ModuleEntry *findModuleEntryNamed(char *name) {
   ModuleEntry *temp = firstEntry;

   while(temp) {
      if(strcmp(temp->name, name) == 0) return temp;
      temp = temp->next;
   }
   return NULL;
}

/* Try to load a given library. Return it's handle or NULL */
HMODULE tryLoading(char *libName) {
   char loadError[255];
   APIRET rc;
   HMODULE h;

   rc = DosLoadModule(loadError, sizeof(loadError), libName, &h);
   if(rc != NO_ERROR) {
      #ifdef WARNINGS
         sprintf(ErrorMessage, "DosLoadModule failed with return code %u for %s\n"
                               "when trying to load module %s.", rc, loadError, libName);
         ShowWarningMessage(ErrorMessage);
      #endif
      h = 0;
   }
   return h;
}

/* Return the module entry for the given module name */
ModuleEntry *loadModuleEntry(char *name) {
   ModuleEntry *temp;
   int ok;

   temp = (ModuleEntry*) calloc(1, sizeof(ModuleEntry));
   strcpy(temp->name, name);
   temp->handle = tryLoading(name);
   if(temp->handle) {
      void *fn;
      APIRET rc;
      /* looks ok so far; fetch setInterpreter() function */
      rc = DosQueryProcAddr(temp->handle, 0, "setInterpreter", (PFN*)&fn);
      if(rc != NO_ERROR) {
         #ifdef WARNINGS
            sprintf(ErrorMessage, "DosQueryProcAddr failed with return code %u\n"
                                  "when trying to find function setInterpreter().", rc);
            ShowWarningMessage(ErrorMessage);
         #endif
         fn = (void*) 0;
      }
      if(fn) {
         struct VirtualMachine *vm;
         /* get the interpreter proxy */
         vm = sqGetInterpreterProxy();
         /* and call setInterpreter() */
         ok = ((int (*) (struct VirtualMachine *)) fn)(vm);
         if(ok) return temp; /* fine; this worked */
      }
   }
   /* something went wrong - clean up stuff */
   if(temp->handle)
      DosFreeModule(temp->handle);
   free(temp);
   return NULL;
}

int ioLoadExternalFunctionOfLengthFromModuleOfLength(int functionNameIndex, int functionNameLength,
                                                     int moduleNameIndex,   int moduleNameLength) {
   int i;
   ModuleEntry *temp;
   char functionName[255];
   void *fn;
   APIRET rc;

   /* Get function name */
   for(i=0; i<functionNameLength; i++)
      functionName[i] = ((char *)functionNameIndex)[i];
   functionName[functionNameLength] = 0;

   /* If in an external module (DLL) */
   if(moduleNameLength != 0) {
      char moduleName[255];

      for(i=0; i<moduleNameLength; i++)
         moduleName[i] = ((unsigned char *)moduleNameIndex)[i];
      moduleName[moduleNameLength] = 0;

      temp = findModuleEntryNamed(moduleName);
      if(!temp) {
         /* new module */
         temp = loadModuleEntry(moduleName);
         if(!temp) return success(false);
         temp->next = firstEntry;
         firstEntry = temp;
      }

      /* Load the function address from the module handle */
      rc = DosQueryProcAddr(temp->handle, 0, functionName, (PFN*)&fn);
      if(rc != NO_ERROR) {
         #ifdef WARNINGS
            sprintf(ErrorMessage, "DosQueryProcAddr failed with return code %u\n"
                                  "when trying to find function %s().", rc, functionName);
            ShowWarningMessage(ErrorMessage);
         #endif
         fn = (void*) 0;
      }

   } else
      /* If in Squeak EXE module */
      fn = getExternalFunctionInSqueakModule(functionName);

   if(!fn)
      return success(false);

   return (int) fn;
}

void* getExternalFunctionInSqueakModule(char* functionName) {
   /* OS/2 only seems to allow finding functions by name inside DLL's, but not
      inside EXE's. One (quite ugly) solution is to manually include here
      all the EXPORTed functions (and copy the declarations at the beginning
      of this file)
      If you find a better way for doing this, please tell me at jvuletic@dc.uba.ar
   */
   int i;
   void *fn = (void*) 0;

   /* Arrays with names and addresses of all EXPORTed functions */
   char* functionNames[] = {
      "gePrimitiveAbortProcessing"          ,
      "gePrimitiveAddActiveEdgeEntry"       ,
      "gePrimitiveAddBezier"                ,
      "gePrimitiveAddBezierShape"           ,
      "gePrimitiveAddBitmapFill"            ,
      "gePrimitiveAddCompressedShape"       ,
      "gePrimitiveAddGradientFill"          ,
      "gePrimitiveAddLine"                  ,
      "gePrimitiveAddOval"                  ,
      "gePrimitiveAddPolygon"               ,
      "gePrimitiveAddRect"                  ,
      "gePrimitiveChangedActiveEdgeEntry"   ,
      "gePrimitiveCopyBuffer"               ,
      "gePrimitiveDisplaySpanBuffer"        ,
      "gePrimitiveDoProfileStats"           ,
      "gePrimitiveFinishedProcessing"       ,
      "gePrimitiveGetAALevel"               ,
      "gePrimitiveGetBezierStats"           ,
      "gePrimitiveGetClipRect"              ,
      "gePrimitiveGetCounts"                ,
      "gePrimitiveGetDepth"                 ,
      "gePrimitiveGetFailureReason"         ,
      "gePrimitiveGetOffset"                ,
      "gePrimitiveGetTimes"                 ,
      "gePrimitiveInitializeBuffer"         ,
      "gePrimitiveInitializeProcessing"     ,
      "gePrimitiveMergeFillFrom"            ,
      "gePrimitiveNeedsFlush"               ,
      "gePrimitiveNeedsFlushPut"            ,
      "gePrimitiveNextActiveEdgeEntry"      ,
      "gePrimitiveNextFillEntry"            ,
      "gePrimitiveNextGlobalEdgeEntry"      ,
      "gePrimitiveRegisterExternalEdge"     ,
      "gePrimitiveRegisterExternalFill"     ,
      "gePrimitiveRenderImage"              ,
      "gePrimitiveRenderScanline"           ,
      "gePrimitiveSetAALevel"               ,
      "gePrimitiveSetClipRect"              ,
      "gePrimitiveSetColorTransform"        ,
      "gePrimitiveSetDepth"                 ,
      "gePrimitiveSetEdgeTransform"         ,
      "gePrimitiveSetOffset"                ,
      "m23PrimitiveComposeMatrix"           ,
      "m23PrimitiveInvertPoint"             ,
      "m23PrimitiveInvertRectInto"          ,
      "m23PrimitiveIsIdentity"              ,
      "m23PrimitiveIsPureTranslation"       ,
      "m23PrimitiveTransformPoint"          ,
      "m23PrimitiveTransformRectInto"       ,
      "primitiveFFTPermuteData"             ,
      "primitiveFFTScaleData"               ,
      "primitiveFFTTransformData"           ,
      "primitiveFloatArrayAddFloatArray"    ,
      "primitiveFloatArrayAddScalar"        ,
      "primitiveFloatArrayAt"               ,
      "primitiveFloatArrayAtPut"            ,
      "primitiveFloatArrayDivFloatArray"    ,
      "primitiveFloatArrayDivScalar"        ,
      "primitiveFloatArrayEqual"            ,
      "primitiveFloatArrayHash"             ,
      "primitiveFloatArrayMulFloatArray"    ,
      "primitiveFloatArrayMulScalar"        ,
      "primitiveFloatArraySubFloatArray"    ,
      "primitiveFloatArraySubScalar"        ,
      "setInterpreter"                      ,
      ""
   };

   void* functionPointer[] = {
       (void*)&gePrimitiveAbortProcessing          ,
       (void*)&gePrimitiveAddActiveEdgeEntry       ,
       (void*)&gePrimitiveAddBezier                ,
       (void*)&gePrimitiveAddBezierShape           ,
       (void*)&gePrimitiveAddBitmapFill            ,
       (void*)&gePrimitiveAddCompressedShape       ,
       (void*)&gePrimitiveAddGradientFill          ,
       (void*)&gePrimitiveAddLine                  ,
       (void*)&gePrimitiveAddOval                  ,
       (void*)&gePrimitiveAddPolygon               ,
       (void*)&gePrimitiveAddRect                  ,
       (void*)&gePrimitiveChangedActiveEdgeEntry   ,
       (void*)&gePrimitiveCopyBuffer               ,
       (void*)&gePrimitiveDisplaySpanBuffer        ,
       (void*)&gePrimitiveDoProfileStats           ,
       (void*)&gePrimitiveFinishedProcessing       ,
       (void*)&gePrimitiveGetAALevel               ,
       (void*)&gePrimitiveGetBezierStats           ,
       (void*)&gePrimitiveGetClipRect              ,
       (void*)&gePrimitiveGetCounts                ,
       (void*)&gePrimitiveGetDepth                 ,
       (void*)&gePrimitiveGetFailureReason         ,
       (void*)&gePrimitiveGetOffset                ,
       (void*)&gePrimitiveGetTimes                 ,
       (void*)&gePrimitiveInitializeBuffer         ,
       (void*)&gePrimitiveInitializeProcessing     ,
       (void*)&gePrimitiveMergeFillFrom            ,
       (void*)&gePrimitiveNeedsFlush               ,
       (void*)&gePrimitiveNeedsFlushPut            ,
       (void*)&gePrimitiveNextActiveEdgeEntry      ,
       (void*)&gePrimitiveNextFillEntry            ,
       (void*)&gePrimitiveNextGlobalEdgeEntry      ,
       (void*)&gePrimitiveRegisterExternalEdge     ,
       (void*)&gePrimitiveRegisterExternalFill     ,
       (void*)&gePrimitiveRenderImage              ,
       (void*)&gePrimitiveRenderScanline           ,
       (void*)&gePrimitiveSetAALevel               ,
       (void*)&gePrimitiveSetClipRect              ,
       (void*)&gePrimitiveSetColorTransform        ,
       (void*)&gePrimitiveSetDepth                 ,
       (void*)&gePrimitiveSetEdgeTransform         ,
       (void*)&gePrimitiveSetOffset                ,
       (void*)&m23PrimitiveComposeMatrix           ,
       (void*)&m23PrimitiveInvertPoint             ,
       (void*)&m23PrimitiveInvertRectInto          ,
       (void*)&m23PrimitiveIsIdentity              ,
       (void*)&m23PrimitiveIsPureTranslation       ,
       (void*)&m23PrimitiveTransformPoint          ,
       (void*)&m23PrimitiveTransformRectInto       ,
       (void*)&primitiveFFTPermuteData             ,
       (void*)&primitiveFFTScaleData               ,
       (void*)&primitiveFFTTransformData           ,
       (void*)&primitiveFloatArrayAddFloatArray    ,
       (void*)&primitiveFloatArrayAddScalar        ,
       (void*)&primitiveFloatArrayAt               ,
       (void*)&primitiveFloatArrayAtPut            ,
       (void*)&primitiveFloatArrayDivFloatArray    ,
       (void*)&primitiveFloatArrayDivScalar        ,
       (void*)&primitiveFloatArrayEqual            ,
       (void*)&primitiveFloatArrayHash             ,
       (void*)&primitiveFloatArrayMulFloatArray    ,
       (void*)&primitiveFloatArrayMulScalar        ,
       (void*)&primitiveFloatArraySubFloatArray    ,
       (void*)&primitiveFloatArraySubScalar        ,
       (void*)&setInterpreter                      ,
       (void*)0
   };

   /* Load the function address from the module handle */
   for(i=0;functionPointer[i];i++)
      if (!strcmp(functionName, functionNames[i])) {
          fn = functionPointer[i];
          break;
       }
   if(!fn) {
      #ifdef WARNINGS
         sprintf(ErrorMessage, "The EXPORTed function %s\n"
                               "included in the Squeak VM was not added to\n"
                               "sqOS2ExternalPrims.c. The primitive will fail.", functionName);
         ShowWarningMessage(ErrorMessage);
      #endif
   }
   return fn;
}
